from tempfile import TemporaryFile
from typing import Tuple, List, Dict

from pulp import LpVariable, LpProblem, LpMaximize, LpMinimize, lpSum, GLPK
from collections import defaultdict
import re
import subprocess


class SolverUtils:

    @staticmethod
    def calculate_solved_problem(
            performance_table_list: List[List[float]],
            preferences: List[List[int]],
            indifferences: List[List[int]],
            criteria: List[bool],
            worst_best_position: List[List[int]],
            number_of_points: List[int],
            comprehensive_intensities: List[List[int]],
            alternative_id_1: int = -1,
            alternative_id_2: int = -1,
            show_logs: bool = False,
    ) -> LpProblem:
        """
        Main calculation method for problem-solving.
        The idea is that this should be a generic method used across different problems

        :param comprehensive_intensities:
        :param performance_table_list:
        :param preferences:
        :param indifferences:
        :param criteria:
        :param worst_best_position:
        :param number_of_points:
        :param alternative_id_1: used only in calculation for hasse graphs
        :param alternative_id_2: used only in calculation for hasse graphs
        :param show_logs: default None

        :return problem:
        """
        problem: LpProblem = LpProblem("UTA-GMS", LpMaximize)

        epsilon: LpVariable = LpVariable("epsilon")

        u_list, u_list_dict = SolverUtils.create_variables_list_and_dict(performance_table_list)

        characteristic_points: List[List[float]] = SolverUtils.calculate_characteristic_points(
            number_of_points, performance_table_list, u_list_dict, u_list
        )

        u_list = [sorted(lp_var_list, key=lambda var: -float(var.name.split("_")[-1]) if len(var.name.split("_")) == 4 else float(var.name.split("_")[-1])) for lp_var_list in u_list]

        # Normalization constraints
        the_greatest_performance: List[LpVariable] = []
        for i in range(len(u_list)):
            if criteria[i]:
                the_greatest_performance.append(u_list[i][-1])
                problem += u_list[i][0] == 0
            else:
                the_greatest_performance.append(u_list[i][0])
                problem += u_list[i][-1] == 0

        problem += lpSum(the_greatest_performance) == 1

        u_list_of_characteristic_points: List[List[LpVariable]] = []
        for i in range(len(characteristic_points)):
            pom = []
            for j in range(len(characteristic_points[i])):
                pom.append(u_list_dict[i][float(characteristic_points[i][j])])
            u_list_of_characteristic_points.append(pom[:])

        # Monotonicity constraint
        for i in range(len(u_list_of_characteristic_points)):
            for j in range(1, len(u_list_of_characteristic_points[i])):
                if criteria[i]:
                    problem += u_list_of_characteristic_points[i][j] >= u_list_of_characteristic_points[i][j - 1]
                else:
                    problem += u_list_of_characteristic_points[i][j - 1] >= u_list_of_characteristic_points[i][j]

        # Bounds constraint
        for i in range(len(u_list_of_characteristic_points)):
            for j in range(1, len(u_list_of_characteristic_points[i]) - 1):
                if criteria[i]:
                    problem += u_list_of_characteristic_points[i][-1] >= u_list_of_characteristic_points[i][j]
                    problem += u_list_of_characteristic_points[i][j] >= u_list_of_characteristic_points[i][0]
                else:
                    problem += u_list_of_characteristic_points[i][0] >= u_list_of_characteristic_points[i][j]
                    problem += u_list_of_characteristic_points[i][j] >= u_list_of_characteristic_points[i][-1]

        # Preference constraint
        for preference in preferences:
            left_alternative: List[float] = performance_table_list[preference[0]]
            right_alternative: List[float] = performance_table_list[preference[1]]

            indices_to_keep: List[int] = preference[2]
            if indices_to_keep:
                left_alternative: List[float] = [left_alternative[i] for i in indices_to_keep]
                right_alternative: List[float] = [right_alternative[i] for i in indices_to_keep]
                left_side: List[LpVariable] = []
                right_side: List[LpVariable] = []
                for i in range(len(indices_to_keep)):
                    left_side.append(u_list_dict[indices_to_keep[i]][left_alternative[i]])
                    right_side.append(u_list_dict[indices_to_keep[i]][right_alternative[i]])
            else:
                left_side: List[LpVariable] = []
                right_side: List[LpVariable] = []
                for i in range(len(left_alternative)):
                    left_side.append(u_list_dict[i][left_alternative[i]])
                    right_side.append(u_list_dict[i][right_alternative[i]])

            problem += lpSum(left_side) >= lpSum(right_side) + epsilon

        # Indifference constraint
        for indifference in indifferences:
            left_alternative: List[float] = performance_table_list[indifference[0]]
            right_alternative: List[float] = performance_table_list[indifference[1]]

            indices_to_keep: List[int] = indifference[2]
            if indices_to_keep:
                left_alternative: List[float] = [left_alternative[i] for i in indices_to_keep]
                right_alternative: List[float] = [right_alternative[i] for i in indices_to_keep]
                left_side: List[LpVariable] = []
                right_side: List[LpVariable] = []
                for i in range(len(indices_to_keep)):
                    left_side.append(u_list_dict[indices_to_keep[i]][left_alternative[i]])
                    right_side.append(u_list_dict[indices_to_keep[i]][right_alternative[i]])
            else:
                left_side: List[LpVariable] = []
                right_side: List[LpVariable] = []
                for i in range(len(left_alternative)):
                    left_side.append(u_list_dict[i][left_alternative[i]])
                    right_side.append(u_list_dict[i][right_alternative[i]])

            problem += lpSum(left_side) == lpSum(right_side)

        if alternative_id_1 >= 0 and alternative_id_2 >= 0:
            left_alternative: List[float] = performance_table_list[alternative_id_2]
            right_alternative: List[float] = performance_table_list[alternative_id_1]

            left_side: List[LpVariable] = []
            right_side: List[LpVariable] = []
            for i in range(len(u_list_dict)):
                left_side.append(u_list_dict[i][left_alternative[i]])
                right_side.append(u_list_dict[i][right_alternative[i]])

            problem += lpSum(left_side) >= lpSum(right_side) + epsilon

        # Worst and Best position
        alternatives_variables: List[List[LpVariable]] = []
        for i in range(len(performance_table_list)):
            pom = []
            for j in range(len(u_list_dict)):
                pom.append(u_list_dict[j][performance_table_list[i][j]])
            alternatives_variables.append(pom[:])

        alternatives_binary_variables: Dict[int, List[Dict[int, LpVariable]]] = {}
        all_binary_variables = {}
        for i in worst_best_position:
            pom_dict = {}
            for j in range(len(performance_table_list)):
                pom = []
                if i[0] != j:
                    variable_1_name: str = f"v_{i[0]}_higher_than_{j}_criteria_{'_'.join(map(str, i[3]))}"
                    if variable_1_name not in all_binary_variables:
                        variable_1: LpVariable = LpVariable(variable_1_name, cat='Binary')
                        pom.append(variable_1)
                        all_binary_variables[variable_1_name] = variable_1
                    else:
                        pom.append(all_binary_variables[variable_1_name])

                    variable_2_name: str = f"v_{j}_higher_than_{i[0]}_criteria_{'_'.join(map(str, i[3]))}"
                    if variable_2_name not in all_binary_variables:
                        variable_2: LpVariable = LpVariable(variable_2_name, cat='Binary')
                        pom.append(variable_2)
                        all_binary_variables[variable_2_name] = variable_2
                    else:
                        pom.append(all_binary_variables[variable_2_name])

                    pom_dict[j] = pom[:]

            if i[0] not in alternatives_binary_variables:
                alternatives_binary_variables[i[0]] = []

            alternatives_binary_variables[i[0]].append(pom_dict)

        big_M: int = 1e20
        x: int = 0
        for worst_best in worst_best_position:
            if len(alternatives_binary_variables[worst_best[0]]) == 1:
                x: int = 0
            for i in range(len(performance_table_list)):
                if i != worst_best[0]:
                    position_constraints: List[LpVariable] = alternatives_variables[worst_best[0]]
                    compared_constraints: List[LpVariable] = alternatives_variables[i]

                    indices_to_keep: List[int] = worst_best[3]
                    if indices_to_keep:
                        position_constraints: List[LpVariable] = [position_constraints[i] for i in indices_to_keep]
                        compared_constraints: List[LpVariable] = [compared_constraints[i] for i in indices_to_keep]

                    problem += lpSum(position_constraints) - lpSum(compared_constraints) + big_M * alternatives_binary_variables[worst_best[0]][x][i][0] >= epsilon

                    problem += lpSum(compared_constraints) - lpSum(position_constraints) + big_M * alternatives_binary_variables[worst_best[0]][x][i][1] >= 0

                    problem += alternatives_binary_variables[worst_best[0]][x][i][0] + alternatives_binary_variables[worst_best[0]][x][i][1] <= 1

            pom_higher = []
            pom_lower = []
            for j in alternatives_binary_variables[worst_best[0]][x]:
                pom_higher.append(alternatives_binary_variables[worst_best[0]][x][j][0])
                pom_lower.append(alternatives_binary_variables[worst_best[0]][x][j][1])
            problem += lpSum(pom_higher) <= worst_best[1] - 1
            problem += lpSum(pom_lower) <= len(performance_table_list) - worst_best[2]

            # If there are more Positions than one, relevant to one alternative
            if len(alternatives_binary_variables[worst_best[0]]) > 1:
                x += 1

        # Use linear interpolation to create constraints
        for i in range(len(u_list_of_characteristic_points)):
            for j in u_list_dict[i]:
                if_characteristic = 0

                for z in range(len(u_list_of_characteristic_points[i])):
                    if u_list_dict[i][j].name == u_list_of_characteristic_points[i][z].name:
                        if_characteristic = 1
                        break

                if if_characteristic == 0:
                    point_before = 0
                    point_after = 1

                    if len(u_list_dict[i][j].name.split("_")) == 4:
                        val = -float(u_list_dict[i][j].name.split("_")[-1])
                    else:
                        val = float(u_list_dict[i][j].name.split("_")[-1])
                    while characteristic_points[i][point_before] > val or val > characteristic_points[i][point_after]:
                        point_before += 1
                        point_after += 1
                    value = SolverUtils.linear_interpolation(val, characteristic_points[i][point_before],
                                                             u_list_dict[i][
                                                                 float(characteristic_points[i][point_before])],
                                                             characteristic_points[i][point_after], u_list_dict[i][
                                                                 float(characteristic_points[i][point_after])])

                    problem += u_list_dict[i][j] == value

        # comprehensive comparisons of intensities of preference
        for intensity in comprehensive_intensities:
            left_alternative_1: List[float] = performance_table_list[intensity[0]]
            left_alternative_2: List[float] = performance_table_list[intensity[2]]
            right_alternative_1: List[float] = performance_table_list[intensity[4]]
            right_alternative_2: List[float] = performance_table_list[intensity[6]]

            left_side_1: List[LpVariable] = []
            left_side_2: List[LpVariable] = []
            right_side_1: List[LpVariable] = []
            right_side_2: List[LpVariable] = []

            indices_to_keep: List[List[int]] = [intensity[1], intensity[3], intensity[5], intensity[7]]
            print(indices_to_keep)
            if indices_to_keep[0]:
                left_alternative_1: List[float] = [left_alternative_1[i] for i in indices_to_keep[0]]
                for i in range(len(indices_to_keep[0])):
                    left_side_1.append(u_list_dict[indices_to_keep[0][i]][left_alternative_1[i]])
            else:
                for i in range(len(left_alternative_1)):
                    left_side_1.append(u_list_dict[i][left_alternative_1[i]])

            if indices_to_keep[1]:
                left_alternative_2: List[float] = [left_alternative_2[i] for i in indices_to_keep[1]]
                for i in range(len(indices_to_keep[1])):
                    left_side_2.append(u_list_dict[indices_to_keep[1][i]][left_alternative_2[i]])
            else:
                for i in range(len(left_alternative_2)):
                    left_side_2.append(u_list_dict[i][left_alternative_2[i]])

            if indices_to_keep[2]:
                right_alternative_1: List[float] = [right_alternative_1[i] for i in indices_to_keep[2]]
                for i in range(len(indices_to_keep[2])):
                    right_side_1.append(u_list_dict[indices_to_keep[2][i]][right_alternative_1[i]])
            else:
                for i in range(len(right_alternative_1)):
                    right_side_1.append(u_list_dict[i][right_alternative_1[i]])

            if indices_to_keep[3]:
                right_alternative_2: List[float] = [right_alternative_2[i] for i in indices_to_keep[3]]
                for i in range(len(indices_to_keep[3])):
                    right_side_2.append(u_list_dict[indices_to_keep[3][i]][right_alternative_2[i]])
            else:
                for i in range(len(right_alternative_2)):
                    right_side_2.append(u_list_dict[i][right_alternative_2[i]])

            if intensity[-1] == '>':
                problem += lpSum(left_side_1) - lpSum(left_side_2) >= lpSum(right_side_1) - lpSum(
                    right_side_2) + epsilon
            elif intensity[-1] == '>=':
                problem += lpSum(left_side_1) - lpSum(left_side_2) >= lpSum(right_side_1) - lpSum(right_side_2)
            else:
                problem += lpSum(left_side_1) - lpSum(left_side_2) == lpSum(right_side_1) - lpSum(right_side_2)

        problem += epsilon

        problem.solve(solver=GLPK(msg=show_logs))

        return problem

    @staticmethod
    def calculate_the_most_representative_function(
            performance_table_list: List[List[float]],
            alternatives_id_list: List[str],
            preferences: List[List[int]],
            indifferences: List[List[int]],
            criteria: List[bool],
            worst_best_position: List[List[int]],
            number_of_points: List[int],
            comprehensive_intensities: List[List[int]],
            show_logs: bool = False,
            sampler_path: str = 'files/polyrun-1.1.0-jar-with-dependencies.jar',
            number_of_samples: str = '100'
    ) -> Tuple[LpProblem, Dict[str, List[int]]]:
        """
        Main method used in getting the most representative value function.

        :param comprehensive_intensities:
        :param performance_table_list:
        :param alternatives_id_list:
        :param preferences:
        :param indifferences:
        :param criteria:
        :param worst_best_position:
        :param number_of_points:
        :param show_logs: default None
        :param sampler_path:
        :param number_of_samples:

        :return problem:
        """
        problem: LpProblem = LpProblem("UTA-GMS", LpMaximize)

        epsilon: LpVariable = LpVariable("epsilon")

        delta: LpVariable = LpVariable("delta")

        u_list, u_list_dict = SolverUtils.create_variables_list_and_dict(performance_table_list)

        characteristic_points: List[List[float]] = SolverUtils.calculate_characteristic_points(
            number_of_points, performance_table_list, u_list_dict, u_list
        )

        u_list = [sorted(lp_var_list, key=lambda var: -float(var.name.split("_")[-1]) if len(var.name.split("_")) == 4 else float(var.name.split("_")[-1])) for lp_var_list in u_list]

        u_list_of_characteristic_points: List[List[LpVariable]] = []
        for i in range(len(characteristic_points)):
            pom = []
            for j in range(len(characteristic_points[i])):
                pom.append(u_list_dict[i][float(characteristic_points[i][j])])
            u_list_of_characteristic_points.append(pom[:])

        # Normalization constraints
        the_greatest_performance: List[LpVariable] = []
        for i in range(len(u_list)):
            if criteria[i]:
                the_greatest_performance.append(u_list[i][-1])
                problem += u_list[i][0] == 0
            else:
                the_greatest_performance.append(u_list[i][0])
                problem += u_list[i][-1] == 0

        problem += lpSum(the_greatest_performance) == 1

        # Monotonicity constraint
        for i in range(len(u_list_of_characteristic_points)):
            for j in range(1, len(u_list_of_characteristic_points[i])):
                if criteria[i]:
                    problem += u_list_of_characteristic_points[i][j] >= u_list_of_characteristic_points[i][j - 1]
                else:
                    problem += u_list_of_characteristic_points[i][j - 1] >= u_list_of_characteristic_points[i][j]

        # Bounds constraint
        for i in range(len(u_list_of_characteristic_points)):
            for j in range(1, len(u_list_of_characteristic_points[i]) - 1):
                if criteria[i]:
                    problem += u_list_of_characteristic_points[i][-1] >= u_list_of_characteristic_points[i][j]
                    problem += u_list_of_characteristic_points[i][j] >= u_list_of_characteristic_points[i][0]
                else:
                    problem += u_list_of_characteristic_points[i][0] >= u_list_of_characteristic_points[i][j]
                    problem += u_list_of_characteristic_points[i][j] >= u_list_of_characteristic_points[i][-1]

        # Preference constraint
        for preference in preferences:
            left_alternative: List[float] = performance_table_list[preference[0]]
            right_alternative: List[float] = performance_table_list[preference[1]]

            indices_to_keep: List[int] = preference[2]
            if indices_to_keep:
                left_alternative: List[float] = [left_alternative[i] for i in indices_to_keep]
                right_alternative: List[float] = [right_alternative[i] for i in indices_to_keep]
                left_side: List[LpVariable] = []
                right_side: List[LpVariable] = []
                for i in range(len(indices_to_keep)):
                    left_side.append(u_list_dict[indices_to_keep[i]][left_alternative[i]])
                    right_side.append(u_list_dict[indices_to_keep[i]][right_alternative[i]])
            else:
                left_side: List[LpVariable] = []
                right_side: List[LpVariable] = []
                for i in range(len(left_alternative)):
                    left_side.append(u_list_dict[i][left_alternative[i]])
                    right_side.append(u_list_dict[i][right_alternative[i]])

            problem += lpSum(left_side) >= lpSum(right_side) + epsilon

        # Indifference constraint
        for indifference in indifferences:
            left_alternative: List[float] = performance_table_list[indifference[0]]
            right_alternative: List[float] = performance_table_list[indifference[1]]

            indices_to_keep: List[int] = indifference[2]
            if indices_to_keep:
                left_alternative: List[float] = [left_alternative[i] for i in indices_to_keep]
                right_alternative: List[float] = [right_alternative[i] for i in indices_to_keep]
                left_side: List[LpVariable] = []
                right_side: List[LpVariable] = []
                for i in range(len(indices_to_keep)):
                    left_side.append(u_list_dict[indices_to_keep[i]][left_alternative[i]])
                    right_side.append(u_list_dict[indices_to_keep[i]][right_alternative[i]])
            else:
                left_side: List[LpVariable] = []
                right_side: List[LpVariable] = []
                for i in range(len(left_alternative)):
                    left_side.append(u_list_dict[i][left_alternative[i]])
                    right_side.append(u_list_dict[i][right_alternative[i]])

            problem += lpSum(left_side) == lpSum(right_side)

        sampler_metrics: Dict[str, List[int]] = SolverUtils.get_sampler_metrics(
            problem=problem,
            performance_table_list=performance_table_list,
            alternatives_id_list=alternatives_id_list,
            sampler_path=sampler_path,
            number_of_samples=number_of_samples,
            u_list_of_characteristic_points=u_list_of_characteristic_points,
            u_list_dict=u_list_dict,
            characteristic_points=characteristic_points
        )

        # Use linear interpolation to create constraints
        for i in range(len(u_list_of_characteristic_points)):
            for j in u_list_dict[i]:
                if_characteristic = 0

                for z in range(len(u_list_of_characteristic_points[i])):
                    if u_list_dict[i][j].name == u_list_of_characteristic_points[i][z].name:
                        if_characteristic = 1
                        break

                if if_characteristic == 0:
                    point_before = 0
                    point_after = 1

                    if len(u_list_dict[i][j].name.split("_")) == 4:
                        val = -float(u_list_dict[i][j].name.split("_")[-1])
                    else:
                        val = float(u_list_dict[i][j].name.split("_")[-1])
                    while characteristic_points[i][point_before] > val or val > characteristic_points[i][point_after]:
                        point_before += 1
                        point_after += 1
                    value = SolverUtils.linear_interpolation(val, characteristic_points[i][point_before], u_list_dict[i][float(characteristic_points[i][point_before])], characteristic_points[i][point_after], u_list_dict[i][float(characteristic_points[i][point_after])])

                    problem += u_list_dict[i][j] == value

        necessary_preference: Dict[str, List[str]] = SolverUtils.get_necessary_relations(
            performance_table_list=performance_table_list,
            alternatives_id_list=alternatives_id_list,
            preferences=preferences,
            indifferences=indifferences,
            criteria=criteria,
            worst_best_position=worst_best_position,
            number_of_points=number_of_points,
            comprehensive_intensities=comprehensive_intensities
        )

        for i in range(len(alternatives_id_list) - 1):
            for j in range(i + 1, len(alternatives_id_list)):
                name_i = alternatives_id_list[i]
                name_j = alternatives_id_list[j]
                pom1 = []
                pom2 = []
                for k in range(len(performance_table_list[i])):
                    pom1.append(u_list_dict[k][float(performance_table_list[i][k])])
                    pom2.append(u_list_dict[k][float(performance_table_list[j][k])])
                sum_i = lpSum(pom1[:])
                sum_j = lpSum(pom2[:])

                if (name_i not in necessary_preference and name_j in necessary_preference and name_i in
                    necessary_preference[name_j]) or \
                        (name_i in necessary_preference and name_j in necessary_preference and name_i in
                         necessary_preference[name_j] and name_j not in necessary_preference[name_i]):
                    problem += sum_j >= sum_i + epsilon
                elif (name_j not in necessary_preference and name_i in necessary_preference and name_j in
                      necessary_preference[name_i]) or \
                        (name_i in necessary_preference and name_j in necessary_preference and name_j in
                         necessary_preference[name_i] and name_i not in necessary_preference[name_j]):
                    problem += sum_i >= sum_j + epsilon
                elif (name_i not in necessary_preference and name_j not in necessary_preference) or \
                        (name_i not in necessary_preference and name_j in necessary_preference and name_i not in
                         necessary_preference[name_j]) or \
                        (name_j not in necessary_preference and name_i in necessary_preference and name_j not in
                         necessary_preference[name_i]) or \
                        (name_i in necessary_preference and name_j not in necessary_preference[
                            name_i] and name_j in necessary_preference and name_i not in necessary_preference[name_j]):
                    problem += sum_i <= delta + sum_j
                    problem += sum_j <= delta + sum_i

        # Worst and Best position
        alternatives_variables: List[List[LpVariable]] = []
        for i in range(len(performance_table_list)):
            pom = []
            for j in range(len(u_list_dict)):
                pom.append(u_list_dict[j][performance_table_list[i][j]])
            alternatives_variables.append(pom[:])

        alternatives_binary_variables: Dict[int, List[Dict[int, LpVariable]]] = {}
        all_binary_variables = {}
        for i in worst_best_position:
            pom_dict = {}
            for j in range(len(performance_table_list)):
                pom = []
                if i[0] != j:
                    variable_1_name: str = f"v_{i[0]}_higher_than_{j}_criteria_{'_'.join(map(str, i[3]))}"
                    if variable_1_name not in all_binary_variables:
                        variable_1: LpVariable = LpVariable(variable_1_name, cat='Binary')
                        pom.append(variable_1)
                        all_binary_variables[variable_1_name] = variable_1
                    else:
                        pom.append(all_binary_variables[variable_1_name])

                    variable_2_name: str = f"v_{j}_higher_than_{i[0]}_criteria_{'_'.join(map(str, i[3]))}"
                    if variable_2_name not in all_binary_variables:
                        variable_2: LpVariable = LpVariable(variable_2_name, cat='Binary')
                        pom.append(variable_2)
                        all_binary_variables[variable_2_name] = variable_2
                    else:
                        pom.append(all_binary_variables[variable_2_name])

                    pom_dict[j] = pom[:]

            if i[0] not in alternatives_binary_variables:
                alternatives_binary_variables[i[0]] = []

            alternatives_binary_variables[i[0]].append(pom_dict)

        big_M: int = 1e20
        x: int = 0
        for worst_best in worst_best_position:
            if len(alternatives_binary_variables[worst_best[0]]) == 1:
                x: int = 0
            for i in range(len(performance_table_list)):
                if i != worst_best[0]:
                    position_constraints: List[LpVariable] = alternatives_variables[worst_best[0]]
                    compared_constraints: List[LpVariable] = alternatives_variables[i]

                    indices_to_keep: List[int] = worst_best[3]
                    if indices_to_keep:
                        position_constraints: List[LpVariable] = [position_constraints[i] for i in indices_to_keep]
                        compared_constraints: List[LpVariable] = [compared_constraints[i] for i in indices_to_keep]

                    problem += lpSum(position_constraints) - lpSum(compared_constraints) + big_M * alternatives_binary_variables[worst_best[0]][x][i][0] >= epsilon

                    problem += lpSum(compared_constraints) - lpSum(position_constraints) + big_M * alternatives_binary_variables[worst_best[0]][x][i][1] >= 0

                    problem += alternatives_binary_variables[worst_best[0]][x][i][0] + alternatives_binary_variables[worst_best[0]][x][i][1] <= 1

            pom_higher = []
            pom_lower = []
            for j in alternatives_binary_variables[worst_best[0]][x]:
                pom_higher.append(alternatives_binary_variables[worst_best[0]][x][j][0])
                pom_lower.append(alternatives_binary_variables[worst_best[0]][x][j][1])
            problem += lpSum(pom_higher) <= worst_best[1] - 1
            problem += lpSum(pom_lower) <= len(performance_table_list) - worst_best[2]

            # If there are more Positions than one, relevant to one alternative
            if len(alternatives_binary_variables[worst_best[0]]) > 1:
                x += 1

        # comprehensive comparisons of intensities of preference
        for intensity in comprehensive_intensities:
            left_alternative_1: List[float] = performance_table_list[intensity[0]]
            left_alternative_2: List[float] = performance_table_list[intensity[2]]
            right_alternative_1: List[float] = performance_table_list[intensity[4]]
            right_alternative_2: List[float] = performance_table_list[intensity[6]]

            left_side_1: List[LpVariable] = []
            left_side_2: List[LpVariable] = []
            right_side_1: List[LpVariable] = []
            right_side_2: List[LpVariable] = []

            indices_to_keep: List[List[int]] = [intensity[1], intensity[3], intensity[5], intensity[7]]
            print(indices_to_keep)
            if indices_to_keep[0]:
                left_alternative_1: List[float] = [left_alternative_1[i] for i in indices_to_keep[0]]
                for i in range(len(indices_to_keep[0])):
                    left_side_1.append(u_list_dict[indices_to_keep[0][i]][left_alternative_1[i]])
            else:
                for i in range(len(left_alternative_1)):
                    left_side_1.append(u_list_dict[i][left_alternative_1[i]])

            if indices_to_keep[1]:
                left_alternative_2: List[float] = [left_alternative_2[i] for i in indices_to_keep[1]]
                for i in range(len(indices_to_keep[1])):
                    left_side_2.append(u_list_dict[indices_to_keep[1][i]][left_alternative_2[i]])
            else:
                for i in range(len(left_alternative_2)):
                    left_side_2.append(u_list_dict[i][left_alternative_2[i]])

            if indices_to_keep[2]:
                right_alternative_1: List[float] = [right_alternative_1[i] for i in indices_to_keep[2]]
                for i in range(len(indices_to_keep[2])):
                    right_side_1.append(u_list_dict[indices_to_keep[2][i]][right_alternative_1[i]])
            else:
                for i in range(len(right_alternative_1)):
                    right_side_1.append(u_list_dict[i][right_alternative_1[i]])

            if indices_to_keep[3]:
                right_alternative_2: List[float] = [right_alternative_2[i] for i in indices_to_keep[3]]
                for i in range(len(indices_to_keep[3])):
                    right_side_2.append(u_list_dict[indices_to_keep[3][i]][right_alternative_2[i]])
            else:
                for i in range(len(right_alternative_2)):
                    right_side_2.append(u_list_dict[i][right_alternative_2[i]])

            if intensity[-1] == '>':
                problem += lpSum(left_side_1) - lpSum(left_side_2) >= lpSum(right_side_1) - lpSum(
                    right_side_2) + epsilon
            elif intensity[-1] == '>=':
                problem += lpSum(left_side_1) - lpSum(left_side_2) >= lpSum(right_side_1) - lpSum(right_side_2)
            else:
                problem += lpSum(left_side_1) - lpSum(left_side_2) == lpSum(right_side_1) - lpSum(right_side_2)

        problem += big_M * epsilon - delta

        problem.solve(solver=GLPK(msg=show_logs))

        return problem, sampler_metrics

    @staticmethod
    def get_necessary_relations(
            performance_table_list: List[List[float]],
            alternatives_id_list: List[str],
            preferences: List[List[int]],
            indifferences: List[List[int]],
            criteria: List[bool],
            worst_best_position: List[List[int]],
            number_of_points: List[int],
            comprehensive_intensities: List[List[int]],
            show_logs: bool = False
    ) -> Dict[str, List[str]]:
        """
        Method used for getting necessary relations.

        :param comprehensive_intensities:
        :param performance_table_list:
        :param alternatives_id_list:
        :param preferences:
        :param indifferences:
        :param criteria:
        :param worst_best_position:
        :param number_of_points:
        :param show_logs: default None

        :return necessary:
        """
        necessary: Dict[str, List[str]] = {}
        for i in range(len(performance_table_list)):
            for j in range(len(performance_table_list)):
                if i == j:
                    continue

                problem: LpProblem = SolverUtils.calculate_solved_problem(
                    performance_table_list=performance_table_list,
                    preferences=preferences,
                    indifferences=indifferences,
                    criteria=criteria,
                    worst_best_position=worst_best_position,
                    number_of_points=number_of_points,
                    comprehensive_intensities=comprehensive_intensities,
                    alternative_id_1=i,
                    alternative_id_2=j,
                    show_logs=show_logs
                )

                if problem.variables()[0].varValue <= 0:
                    if alternatives_id_list[i] not in necessary:
                        necessary[alternatives_id_list[i]] = []
                    necessary[alternatives_id_list[i]].append(alternatives_id_list[j])

        return necessary

    @staticmethod
    def create_variables_list_and_dict(performance_table: List[list]) -> Tuple[List[list], List[dict]]:
        """
        Method responsible for creating a technical list of variables and a technical dict of variables that are used
        for adding constraints to the problem.

        :param performance_table:

        :return u_list, u_list_dict: ex. Tuple([[u_0_0.0, u_0_2.0], [u_1_2.0, u_1_9.0]], [{26.0: u_0_26.0, 2.0: u_0_2.0}, {40.0: u_1_40.0, 2.0: u_1_2.0}])
        """
        u_list: List[List[LpVariable]] = []
        u_list_dict: List[Dict[float, LpVariable]] = []

        for i in range(len(performance_table[0])):
            row: List[LpVariable] = []
            row_dict: Dict[float, LpVariable] = {}

            for j in range(len(performance_table)):
                variable_name: str = f"u_{i}_{float(performance_table[j][i])}"
                variable: LpVariable = LpVariable(variable_name)

                if performance_table[j][i] not in row_dict:
                    row_dict[float(performance_table[j][i])] = variable

                flag: int = 1
                for var in row:
                    if str(var) == variable_name:
                        flag: int = 0
                if flag:
                    row.append(variable)

            u_list_dict.append(row_dict)

            row = sorted(row, key=lambda var: -float(var.name.split("_")[-1]) if len(var.name.split("_")) == 4 else float(var.name.split("_")[-1]))
            u_list.append(row)

        return u_list, u_list_dict

    @staticmethod
    def calculate_direct_relations(necessary: Dict[str, List[str]]) -> Dict[str, List[str]]:
        """
        Method for getting only direct relations in Hasse Diagram
        :param necessary:
        :return direct_relations:
        """
        direct_relations: Dict[str, List[str]] = {}
        # first create the relation list for each node
        for node1, relations in necessary.items():
            direct_relations[node1] = sorted(relations)
        # then prune the indirect relations
        for node1, related_nodes in list(direct_relations.items()):  # make a copy of items
            related_nodes_copy: List[str] = related_nodes.copy()
            for node2 in related_nodes:
                # Check if node2 is also related to any other node that is related to node1
                for other_node in related_nodes:
                    if other_node != node2 and other_node in direct_relations and node2 in direct_relations[other_node]:
                        # If such a relationship exists, remove the relation between node1 and node2
                        related_nodes_copy.remove(node2)
                        break
            direct_relations[node1] = sorted(related_nodes_copy)  # sort the list

        return direct_relations

    @staticmethod
    def get_alternatives_and_utilities_dict(
            variables_and_values_dict,
            performance_table_list,
            alternatives_id_list,
    ) -> Dict[str, float]:
        """
        Method for getting alternatives_and_utilities_dict

        :param variables_and_values_dict:
        :param performance_table_list:
        :param alternatives_id_list:

        :return sorted_dict:
        """

        utilities: List[float] = []
        for i in range(len(performance_table_list)):
            utility: float = 0.0
            for j in range(len(performance_table_list[i])):
                variable_name: str = f"u_{j}_{performance_table_list[i][j]}"
                if '-' in variable_name:
                    variable_name: str = variable_name.replace('-', '_')
                utility += round(variables_and_values_dict[variable_name], 4)

            utilities.append(round(utility, 4))

        utilities_dict: Dict[str, float] = {}
        # TODO: Sorting possibly unnecessary, but for now it's nicer for human eye :)
        for i in range(len(utilities)):
            utilities_dict[alternatives_id_list[i]] = utilities[i]
        sorted_dict: Dict[str, float] = dict(sorted(utilities_dict.items(), key=lambda item: item[1]))

        return sorted_dict

    @staticmethod
    def calculate_characteristic_points(
            number_of_points,
            performance_table_list,
            u_list_dict,
            u_list
    ) -> List[List[float]]:
        """
        Method for calculating characteristic points

        :param number_of_points:
        :param performance_table_list:
        :param u_list_dict:
        :param u_list:

        :return characteristic_points:
        """
        columns: List[Tuple[float]] = list(zip(*performance_table_list))
        worst_values: List[float] = [min(col) for col in columns]
        best_values: List[float] = [max(col) for col in columns]
        characteristic_points: List[List[float]] = []

        for i in range(len(worst_values)):
            pom = []
            if number_of_points[i] != 0:
                for j in range(number_of_points[i]):
                    x = worst_values[i] + (j / (number_of_points[i] - 1)) * (best_values[i] - worst_values[i])
                    if x not in u_list_dict[i]:
                        new: str = f"u_{i}_{x}"
                        variable: LpVariable = LpVariable(new)
                        new: Dict[float, LpVariable] = {x: variable}
                        u_list_dict[i].update(new)
                        u_list[i].append(variable)
                    pom.append(x)
                characteristic_points.append(pom[:])
            else:
                for j in range(len(performance_table_list)):
                    if float(performance_table_list[j][i]) not in pom:
                        pom.append(float(performance_table_list[j][i]))
                pom.sort()
                characteristic_points.append(pom[:])
        return characteristic_points

    @staticmethod
    def linear_interpolation(x, x1, y1, x2, y2) -> float:
        """Perform linear interpolation to estimate a value at a specific point on a straight line"""
        result = y1 + ((x - x1) * (y2 - y1)) / (x2 - x1)
        return result

    @staticmethod
    def get_criterion_functions(
            variables_and_values_dict,
            criteria
    ) -> Dict[str, List[Tuple[float, float]]]:
        """
        Method responsible for getting criterion functions

        :param variables_and_values_dict:
        :param criteria:
        :return:
        """
        criterion_functions: Dict[str, List[Tuple[float, float]]] = defaultdict(list)

        criterion_ids: List[str] = []
        for crit in criteria:
            criterion_ids.append(crit.criterion_id)

        for key, value in variables_and_values_dict.items():
            if key.startswith('u'):
                first_part, x_value = key.rsplit('_', 1)
                if first_part.endswith('_'):
                    _, i, __ = first_part.rsplit('_', 2)
                else:
                    _, i = first_part.rsplit('_', 1)

                criterion_functions[criterion_ids[int(i)]].append((float(x_value), value))

        for key, values in criterion_functions.items():
            criterion_functions[key] = sorted(values, key=lambda x: x[0])

        return dict(criterion_functions)

    @staticmethod
    def get_sampler_metrics(
            problem,
            performance_table_list,
            alternatives_id_list,
            sampler_path,
            number_of_samples,
            u_list_of_characteristic_points,
            u_list_dict,
            characteristic_points
    ) -> Dict[str, List[int]]:
        # Write input file for Sampler
        with TemporaryFile("w+") as input_file, TemporaryFile("w+") as output_file:
            # Write header, useful only for testing
            variable_names = [var.name for var in problem.variables()]

            for constraint in problem.constraints.values():
                constraint_values = []
                for var in problem.variables():
                    if var in constraint:
                        constraint_values.append(str(constraint[var]))
                    else:
                        constraint_values.append("0")
                constraint_values.append(re.search(r'([<>=]=?)', str(constraint)).group(1))
                constraint_values.append(str(-constraint.constant))
                input_file.write(" ".join(constraint_values) + "\n")

            input_file.seek(0)
            # Write Sampler output file
            subprocess.call(
                ['java', '-jar', sampler_path, '-n', number_of_samples],
                stdin=input_file,
                stdout=output_file
            )

            output: Dict[str, List[float]] = {}
            for alternative in alternatives_id_list:
                output[alternative] = [0] * len(alternatives_id_list)

            output_file.seek(0)
            for line in output_file:
                variables_and_values_dict: Dict[str, float] = {}

                if 'epsilon' in variable_names:
                    var_names = variable_names[1:]
                    values = line.strip().split('\t')[1:]
                else:
                    var_names = variable_names
                    values = line.strip().split('\t')

                for var_name, var_value in zip(var_names, values):
                    variables_and_values_dict[var_name] = float(var_value)

                # need to add interpolation here to variables_and_values_dict
                # Use linear interpolation to create constraints
                for i in range(len(u_list_of_characteristic_points)):
                    for j in u_list_dict[i]:
                        if_characteristic = 0

                        for z in range(len(u_list_of_characteristic_points[i])):
                            if u_list_dict[i][j].name == u_list_of_characteristic_points[i][z].name:
                                if_characteristic = 1
                                break

                        if if_characteristic == 0:
                            point_before = 0
                            point_after = 1

                            if len(u_list_dict[i][j].name.split("_")) == 4:
                                val = -float(u_list_dict[i][j].name.split("_")[-1])
                            else:
                                val = float(u_list_dict[i][j].name.split("_")[-1])
                            while characteristic_points[i][point_before] > val or val > characteristic_points[i][point_after]:
                                point_before += 1
                                point_after += 1
                            value = SolverUtils.linear_interpolation(val, characteristic_points[i][point_before], u_list_dict[i][float(characteristic_points[i][point_before])], characteristic_points[i][point_after], u_list_dict[i][float(characteristic_points[i][point_after])])

                            result: float  = sum(coef * variables_and_values_dict[var.name] for var, coef in value.items())
                            variable_name: str = str(u_list_dict[i][j])
                            variables_and_values_dict[variable_name] = result

                alternatives_and_utilities_dict: Dict[str, float] = SolverUtils.get_alternatives_and_utilities_dict(
                    variables_and_values_dict=variables_and_values_dict,
                    performance_table_list=performance_table_list,
                    alternatives_id_list=alternatives_id_list,
                )

                letter_value_pairs = [(letter, value) for letter, value in alternatives_and_utilities_dict.items()]

                letter_value_pairs.sort(key=lambda x: x[1], reverse=True)

                single_ranking = {}
                place = 1
                for i in range(len(letter_value_pairs)):
                    letter, value = letter_value_pairs[i]

                    # Check if the current value is the same as the previous value
                    if i > 0 and value == letter_value_pairs[i - 1][1]:
                        single_ranking[letter] = single_ranking[letter_value_pairs[i - 1][0]]
                    else:
                        single_ranking[letter] = place

                    place += 1

                for key, value in single_ranking.items():
                    output[key][value-1] = output[key][value-1] + 1

            for key, value in output.items():
                try:
                    output[key] = [round(val / sum(output[key]) * 100, 10) for val in value]
                except:
                    output[key] = []

            return output

    @staticmethod
    def resolve_incosistency(
            performance_table_list: List[List[float]],
            preferences: List[List[int]],
            indifferences: List[List[int]],
            criteria: List[bool],
            worst_best_position: List[List[int]],
            number_of_points: List[int],
            comprehensive_intensities: List[List[int]],
            subsets_to_remove: List[List[List[List[int]]]],
            show_logs: bool = False,
    ):
        """
        Main calculation method for problem-solving.
        The idea is that this should be a generic method used across different problems

        :param subsets_to_remove:
        :param performance_table_list:
        :param preferences:
        :param indifferences:
        :param criteria:
        :param worst_best_position:
        :param number_of_points:
        :param comprehensive_intensities:
        :param alternative_id_1: used only in calculation for hasse graphs
        :param alternative_id_2: used only in calculation for hasse graphs
        :param show_logs: default None

        :return problem:
        """
        problem: LpProblem = LpProblem("UTA-GMS", LpMinimize)

        epsilon: LpVariable = LpVariable("epsilon")

        u_list, u_list_dict = SolverUtils.create_variables_list_and_dict(performance_table_list)

        characteristic_points: List[List[float]] = SolverUtils.calculate_characteristic_points(
            number_of_points, performance_table_list, u_list_dict, u_list
        )

        u_list = [sorted(lp_var_list,
                         key=lambda var: -float(var.name.split("_")[-1]) if len(var.name.split("_")) == 4 else float(
                             var.name.split("_")[-1])) for lp_var_list in u_list]

        u_list_of_characteristic_points: List[List[LpVariable]] = []
        for i in range(len(characteristic_points)):
            pom = []
            for j in range(len(characteristic_points[i])):
                pom.append(u_list_dict[i][float(characteristic_points[i][j])])
            u_list_of_characteristic_points.append(pom[:])

        problem += epsilon == 0.0001

        # Normalization constraints
        the_greatest_performance: List[LpVariable] = []

        for i in range(len(u_list)):

            if criteria[i]:
                the_greatest_performance.append(u_list[i][-1])
                problem += u_list[i][0] == 0
            else:
                the_greatest_performance.append(u_list[i][0])
                problem += u_list[i][-1] == 0

        problem += lpSum(the_greatest_performance) == 1

        # Monotonicity constraint
        for i in range(len(u_list_of_characteristic_points)):
            for j in range(1, len(u_list_of_characteristic_points[i])):
                if criteria[i]:
                    problem += u_list_of_characteristic_points[i][j] >= u_list_of_characteristic_points[i][j - 1]
                else:
                    problem += u_list_of_characteristic_points[i][j - 1] >= u_list_of_characteristic_points[i][j]

        # Bounds constraint
        for i in range(len(u_list_of_characteristic_points)):
            for j in range(1, len(u_list_of_characteristic_points[i]) - 1):
                if criteria[i]:
                    problem += u_list_of_characteristic_points[i][-1] >= u_list_of_characteristic_points[i][j]
                    problem += u_list_of_characteristic_points[i][j] >= u_list_of_characteristic_points[i][0]
                else:
                    problem += u_list_of_characteristic_points[i][0] >= u_list_of_characteristic_points[i][j]
                    problem += u_list_of_characteristic_points[i][j] >= u_list_of_characteristic_points[i][-1]

        binary_variables_inconsistency_dict = {}
        binary_variables_inconsistency_list_worst_best = []
        # Worst and Best position
        alternatives_variables: List[List[LpVariable]] = []
        for i in range(len(performance_table_list)):
            pom = []
            for j in range(len(u_list_dict)):
                pom.append(u_list_dict[j][performance_table_list[i][j]])
            alternatives_variables.append(pom[:])

        alternatives_binary_variables: Dict[int, List[Dict[int, LpVariable]]] = {}
        all_binary_variables = {}
        for i in worst_best_position:
            pom_dict = {}
            for j in range(len(performance_table_list)):
                pom = []
                if i[0] != j:
                    variable_1_name: str = f"v_{i[0]}_higher_than_{j}_criteria_{'_'.join(map(str, i[3]))}"
                    if variable_1_name not in all_binary_variables:
                        variable_1: LpVariable = LpVariable(variable_1_name, cat='Binary')
                        pom.append(variable_1)
                        all_binary_variables[variable_1_name] = variable_1
                    else:
                        pom.append(all_binary_variables[variable_1_name])

                    variable_2_name: str = f"v_{j}_higher_than_{i[0]}_criteria_{'_'.join(map(str, i[3]))}"
                    if variable_2_name not in all_binary_variables:
                        variable_2: LpVariable = LpVariable(variable_2_name, cat='Binary')
                        pom.append(variable_2)
                        all_binary_variables[variable_2_name] = variable_2
                    else:
                        pom.append(all_binary_variables[variable_2_name])

                    pom_dict[j] = pom[:]

            if i[0] not in alternatives_binary_variables:
                alternatives_binary_variables[i[0]] = []

            alternatives_binary_variables[i[0]].append(pom_dict)

        big_M: float = 1e20
        x: int = 0
        for worst_best in worst_best_position:
            if len(alternatives_binary_variables[worst_best[0]]) == 1:
                x: int = 0
            for i in range(len(performance_table_list)):
                if i != worst_best[0]:

                    position_constraints: List[LpVariable] = alternatives_variables[worst_best[0]]
                    compared_constraints: List[LpVariable] = alternatives_variables[i]

                    indices_to_keep: List[int] = worst_best[3]
                    if indices_to_keep:
                        position_constraints: List[LpVariable] = [position_constraints[i] for i in indices_to_keep]
                        compared_constraints: List[LpVariable] = [compared_constraints[i] for i in indices_to_keep]

                    variable: str = f"vwb_{worst_best[0]}_{worst_best[1]}_{worst_best[2]}_criteria_{'_'.join(map(str, worst_best[3]))}"
                    if variable not in binary_variables_inconsistency_dict:
                        variable_1: LpVariable = LpVariable(variable, cat='Binary')
                        binary_variables_inconsistency_dict[variable] = variable_1
                        binary_variables_inconsistency_list_worst_best.append(variable_1)

                    problem += lpSum(position_constraints) - lpSum(compared_constraints) + variable_1 * big_M + big_M * \
                               alternatives_binary_variables[worst_best[0]][x][i][0] >= epsilon

                    problem += lpSum(compared_constraints) - lpSum(position_constraints) + variable_1 * big_M + big_M * \
                               alternatives_binary_variables[worst_best[0]][x][i][1] >= 0

                    problem += alternatives_binary_variables[worst_best[0]][x][i][0] + \
                               alternatives_binary_variables[worst_best[0]][x][i][1] <= 1 + variable_1 * big_M

            pom_higher = []
            pom_lower = []
            for j in alternatives_binary_variables[worst_best[0]][x]:
                pom_higher.append(alternatives_binary_variables[worst_best[0]][x][j][0])
                pom_lower.append(alternatives_binary_variables[worst_best[0]][x][j][1])
            problem += lpSum(pom_higher) <= worst_best[1] - 1 + big_M * variable_1
            problem += lpSum(pom_lower) <= len(performance_table_list) - worst_best[2] + big_M * variable_1

            # If there are more Positions than one, relevant to one alternative
            if len(alternatives_binary_variables[worst_best[0]]) > 1:
                x += 1

        binary_variables_inconsistency_list_preferences = []
        # Preference constraint
        for preference in preferences:
            left_alternative: List[float] = performance_table_list[preference[0]]
            right_alternative: List[float] = performance_table_list[preference[1]]

            indices_to_keep: List[int] = preference[2]
            if indices_to_keep:
                left_alternative: List[float] = [left_alternative[i] for i in indices_to_keep]
                right_alternative: List[float] = [right_alternative[i] for i in indices_to_keep]
                left_side: List[LpVariable] = []
                right_side: List[LpVariable] = []
                for i in range(len(indices_to_keep)):
                    left_side.append(u_list_dict[indices_to_keep[i]][left_alternative[i]])
                    right_side.append(u_list_dict[indices_to_keep[i]][right_alternative[i]])
            else:
                left_side: List[LpVariable] = []
                right_side: List[LpVariable] = []
                for i in range(len(left_alternative)):
                    left_side.append(u_list_dict[i][left_alternative[i]])
                    right_side.append(u_list_dict[i][right_alternative[i]])

            variable: str = f"vp_{preference[0]}_{preference[1]}_criteria_{'_'.join(map(str, preference[2]))}"
            if variable not in binary_variables_inconsistency_dict:
                variable_1: LpVariable = LpVariable(variable, cat='Binary')
                binary_variables_inconsistency_dict[variable] = variable_1
                binary_variables_inconsistency_list_preferences.append(variable_1)

            if preference[0] == preference[1]:
                problem += lpSum(left_side) >= lpSum(right_side) + epsilon - big_M * variable_1
                problem += variable_1 == 1
            else:
                problem += lpSum(left_side) >= lpSum(right_side) + epsilon - big_M * variable_1

        binary_variables_inconsistency_list_indifferences = []
        # Indifference constraint
        for indifference in indifferences:
            left_alternative: List[float] = performance_table_list[indifference[0]]
            right_alternative: List[float] = performance_table_list[indifference[1]]

            indices_to_keep: List[int] = indifference[2]
            if indices_to_keep:
                left_alternative: List[float] = [left_alternative[i] for i in indices_to_keep]
                right_alternative: List[float] = [right_alternative[i] for i in indices_to_keep]
                left_side: List[LpVariable] = []
                right_side: List[LpVariable] = []
                for i in range(len(indices_to_keep)):
                    left_side.append(u_list_dict[indices_to_keep[i]][left_alternative[i]])
                    right_side.append(u_list_dict[indices_to_keep[i]][right_alternative[i]])
            else:
                left_side: List[LpVariable] = []
                right_side: List[LpVariable] = []
                for i in range(len(left_alternative)):
                    left_side.append(u_list_dict[i][left_alternative[i]])
                    right_side.append(u_list_dict[i][right_alternative[i]])

            variable: str = f"vi_{indifference[0]}_{indifference[1]}_criteria_{'_'.join(map(str, indifference[2]))}"
            if variable not in binary_variables_inconsistency_dict:
                variable_1: LpVariable = LpVariable(variable, cat='Binary')
                binary_variables_inconsistency_dict[variable] = variable_1
                binary_variables_inconsistency_list_indifferences.append(variable_1)

            problem += lpSum(left_side) + big_M * variable_1 >= lpSum(right_side)
            problem += lpSum(right_side) + big_M * variable_1 >= lpSum(left_side)

        binary_variables_inconsistency_list_comprehensive_intensities = []
        # comprehensive comparisons of intensities of preference
        for intensity in comprehensive_intensities:
            left_alternative_1: List[float] = performance_table_list[intensity[0]]
            left_alternative_2: List[float] = performance_table_list[intensity[2]]
            right_alternative_1: List[float] = performance_table_list[intensity[4]]
            right_alternative_2: List[float] = performance_table_list[intensity[6]]

            left_side_1: List[LpVariable] = []
            left_side_2: List[LpVariable] = []
            right_side_1: List[LpVariable] = []
            right_side_2: List[LpVariable] = []

            indices_to_keep: List[List[int]] = [intensity[1], intensity[3], intensity[5], intensity[7]]

            if indices_to_keep[0]:
                left_alternative_1: List[float] = [left_alternative_1[i] for i in indices_to_keep[0]]
                for i in range(len(indices_to_keep[0])):
                    left_side_1.append(u_list_dict[indices_to_keep[0][i]][left_alternative_1[i]])
            else:
                for i in range(len(left_alternative_1)):
                    left_side_1.append(u_list_dict[i][left_alternative_1[i]])

            if indices_to_keep[1]:
                left_alternative_2: List[float] = [left_alternative_2[i] for i in indices_to_keep[1]]
                for i in range(len(indices_to_keep[1])):
                    left_side_2.append(u_list_dict[indices_to_keep[1][i]][left_alternative_2[i]])
            else:
                for i in range(len(left_alternative_2)):
                    left_side_2.append(u_list_dict[i][left_alternative_2[i]])

            if indices_to_keep[2]:
                right_alternative_1: List[float] = [right_alternative_1[i] for i in indices_to_keep[2]]
                for i in range(len(indices_to_keep[2])):
                    right_side_1.append(u_list_dict[indices_to_keep[2][i]][right_alternative_1[i]])
            else:
                for i in range(len(right_alternative_1)):
                    right_side_1.append(u_list_dict[i][right_alternative_1[i]])

            if indices_to_keep[3]:
                right_alternative_2: List[float] = [right_alternative_2[i] for i in indices_to_keep[3]]
                for i in range(len(indices_to_keep[3])):
                    right_side_2.append(u_list_dict[indices_to_keep[3][i]][right_alternative_2[i]])
            else:
                for i in range(len(right_alternative_2)):
                    right_side_2.append(u_list_dict[i][right_alternative_2[i]])

            if intensity[8] == '=':
                relation = 'e'
            elif intensity[8] == '>':
                relation = 'g'
            elif intensity[8] == '>=':
                relation = 'ge'

            variable: str = f"vci_{intensity[0]}_{intensity[2]}_{intensity[4]}_{intensity[6]}_{relation}_c_{'_'.join(map(str, intensity[1]))}_c_{'_'.join(map(str, intensity[3]))}_c_{'_'.join(map(str, intensity[5]))}_c_{'_'.join(map(str, intensity[7]))}"
            if variable not in binary_variables_inconsistency_dict:
                variable_1: LpVariable = LpVariable(variable, cat='Binary')
                binary_variables_inconsistency_dict[variable] = variable_1
                binary_variables_inconsistency_list_comprehensive_intensities.append(variable_1)

            if intensity[-1] == '>':
                if (intensity[0] == intensity[2] and intensity[1] == intensity[3] and intensity[4] == intensity[6] and
                    intensity[5] == intensity[7]) or (
                        intensity[0] == intensity[4] and intensity[1] == intensity[5] and intensity[2] == intensity[
                    6] and intensity[3] == intensity[7]):
                    problem += lpSum(left_side_1) - lpSum(left_side_2) >= lpSum(right_side_1) - lpSum(
                        right_side_2) + epsilon - big_M * variable_1
                    problem += variable_1 == 1
                else:
                    problem += lpSum(left_side_1) - lpSum(left_side_2) >= lpSum(right_side_1) - lpSum(
                        right_side_2) + epsilon - big_M * variable_1
            elif intensity[-1] == '>=':
                problem += lpSum(left_side_1) - lpSum(left_side_2) + big_M * variable_1 >= lpSum(right_side_1) - lpSum(
                    right_side_2)
            else:
                problem += lpSum(left_side_1) - lpSum(left_side_2) + big_M * variable_1 >= lpSum(right_side_1) - lpSum(
                    right_side_2)
                problem += lpSum(left_side_1) - lpSum(left_side_2) <= lpSum(right_side_1) - lpSum(
                    right_side_2) + big_M * variable_1

        # Use linear interpolation to create constraints
        for i in range(len(u_list_of_characteristic_points)):
            for j in u_list_dict[i]:
                if_characteristic = 0

                for z in range(len(u_list_of_characteristic_points[i])):
                    if u_list_dict[i][j].name == u_list_of_characteristic_points[i][z].name:
                        if_characteristic = 1
                        break

                if if_characteristic == 0:
                    point_before = 0
                    point_after = 1

                    if len(u_list_dict[i][j].name.split("_")) == 4:
                        val = -float(u_list_dict[i][j].name.split("_")[-1])
                    else:
                        val = float(u_list_dict[i][j].name.split("_")[-1])
                    while characteristic_points[i][point_before] > val or val > characteristic_points[i][point_after]:
                        point_before += 1
                        point_after += 1
                    value = SolverUtils.linear_interpolation(val, characteristic_points[i][point_before],
                                                             u_list_dict[i][
                                                                 float(characteristic_points[i][point_before])],
                                                             characteristic_points[i][point_after], u_list_dict[i][
                                                                 float(characteristic_points[i][point_after])])

                    problem += u_list_dict[i][j] == value

        if subsets_to_remove != []:
            for i in range(len(subsets_to_remove)):
                pom = []
                for j in range(len(subsets_to_remove[i])):
                    for k in range(len(subsets_to_remove[i][j])):
                        if j == 0:
                            variable: str = f"vp_{subsets_to_remove[i][j][k][0]}_{subsets_to_remove[i][j][k][1]}_criteria_{'_'.join(map(str, subsets_to_remove[i][j][k][2]))}"
                            pom.append(binary_variables_inconsistency_dict[variable])
                        elif j == 1:
                            variable: str = f"vi_{subsets_to_remove[i][j][k][0]}_{subsets_to_remove[i][j][k][1]}_criteria_{'_'.join(map(str, subsets_to_remove[i][j][k][2]))}"
                            pom.append(binary_variables_inconsistency_dict[variable])
                        elif j == 2:
                            variable: str = f"vwb_{subsets_to_remove[i][j][k][0]}_{subsets_to_remove[i][j][k][1]}_{subsets_to_remove[i][j][k][2]}_criteria_{'_'.join(map(str, subsets_to_remove[i][j][k][3]))}"
                            pom.append(binary_variables_inconsistency_dict[variable])
                        elif j == 3:

                            if subsets_to_remove[i][j][k][8] == '=':
                                relation = 'e'
                            elif subsets_to_remove[i][j][k][8] == '>':
                                relation = 'g'
                            elif subsets_to_remove[i][j][k][8] == '>=':
                                relation = 'ge'

                            variable: str = f"vci_{subsets_to_remove[i][j][k][0]}_{subsets_to_remove[i][j][k][2]}_{subsets_to_remove[i][j][k][4]}_{subsets_to_remove[i][j][k][6]}_{relation}_c_{'_'.join(map(str, subsets_to_remove[i][j][k][1]))}_c_{'_'.join(map(str, subsets_to_remove[i][j][k][3]))}_c_{'_'.join(map(str, subsets_to_remove[i][j][k][5]))}_c_{'_'.join(map(str, subsets_to_remove[i][j][k][7]))}"
                            pom.append(binary_variables_inconsistency_dict[variable])

                problem += lpSum(pom[:]) <= len(subsets_to_remove[i][0]) + len(subsets_to_remove[i][1]) + len(
                    subsets_to_remove[i][2]) + len(subsets_to_remove[i][3]) - 1

        v = lpSum(binary_variables_inconsistency_list_preferences) + lpSum(
            binary_variables_inconsistency_list_indifferences) + lpSum(
            binary_variables_inconsistency_list_worst_best) + lpSum(
            binary_variables_inconsistency_list_comprehensive_intensities)
        problem += v

        problem.solve(solver=GLPK(msg=show_logs))

        result = []
        resultp = []
        resulti = []
        resultwb = []
        resultci = []
        for i in problem.variables():
            pom = []
            name = i.name
            numbers_in_string = name.split("_")
            if i.name[0] == "v" and i.name[1] == "i" and i.varValue == 1:
                criterion = []
                for j in range(1, len(numbers_in_string)):
                    if j == 3:
                        continue
                    elif j > 2:
                        if numbers_in_string[j] == "":
                            continue
                        else:
                            criterion.append(int(numbers_in_string[j]))
                    else:
                        pom.append(int(numbers_in_string[j]))
                pom.append(criterion[:])
                resulti.append(pom[:])
            elif i.name[0] == "v" and i.name[1] == "p" and i.name[2] != 'i' and i.varValue == 1:
                criterion = []
                for j in range(1, len(numbers_in_string)):
                    if j == 3:
                        continue
                    elif j > 2:
                        if numbers_in_string[j] == "":
                            continue
                        else:
                            criterion.append(int(numbers_in_string[j]))
                    else:
                        pom.append(int(numbers_in_string[j]))
                pom.append(criterion[:])
                resultp.append(pom[:])
            elif i.name[0] == "v" and i.name[1] == "w" and i.name[2] == "b" and i.varValue == 1:
                criterion = []
                for j in range(1, len(numbers_in_string)):
                    if j == 4:
                        continue
                    elif j > 4:
                        if numbers_in_string[j] == "":
                            continue
                        else:
                            criterion.append(int(numbers_in_string[j]))
                    else:
                        pom.append(int(numbers_in_string[j]))
                pom.append(criterion[:])
                resultwb.append(pom[:])
            elif i.name[0] == "v" and i.name[1] == "c" and i.name[2] == "i" and i.varValue == 1:
                criterion = []
                for x in range(1, 5):
                    pom.append(int(numbers_in_string[x]))
                if numbers_in_string[5] == 'e':
                    relation = '='
                elif numbers_in_string[5] == 'g':
                    relation = '>'
                elif numbers_in_string[5] == 'ge':
                    relation = '>='
                pom.append(relation)
                x = 7
                pom_criteria = []
                for y in range(4):
                    while numbers_in_string[x] != "c" and numbers_in_string[x] != "":
                        pom_criteria.append(int(numbers_in_string[x]))
                        x = x + 1

                        if x >= len(numbers_in_string):
                            break
                    criterion.append(pom_criteria[:])
                    x = x + 1

                    pom_criteria = []

                pom_final = []
                for x in range(5):
                    if x == 4:
                        pom_final.append(pom[x])
                    else:
                        pom_final.append(pom[x])
                        pom_final.append(criterion[x])
                resultci.append(pom_final[:])

        result.append(resultp)
        result.append(resulti)
        result.append(resultwb)
        result.append(resultci)
        subsets_to_remove.append(result)

        if subsets_to_remove[-1][0] == [] and subsets_to_remove[-1][1] == [] and subsets_to_remove[-1][2] == [] and \
                subsets_to_remove[-1][3] == []:
            return subsets_to_remove
        else:
            return SolverUtils.resolve_incosistency(
                performance_table_list,
                preferences,
                indifferences,
                criteria,
                worst_best_position,
                number_of_points,
                comprehensive_intensities,
                subsets_to_remove
            )



import pytest
from src.utagmsengine.solver import Solver, Inconsistency
from src.utagmsengine.dataclasses import Preference, Indifference, Criterion, Position, Intensity


@pytest.fixture()
def performance_table_dict_dummy():
    return {
        'A': {'g1': 26.0, 'g2': 40.0, 'g3': 44.0},
        'B': {'g1': 2.0, 'g2': 2.0, 'g3': 68.0},
        'C': {'g1': 18.0, 'g2': 17.0, 'g3': 14.0},
        'D': {'g1': 35.0, 'g2': 62.0, 'g3': 25.0},
        'E': {'g1': -7.0, 'g2': 55.0, 'g3': 12.0},
        'F': {'g1': 25.0, 'g2': 30.0, 'g3': 12.0},
        'G': {'g1': 9.0, 'g2': 62.0, 'g3': 88.0},
        'H': {'g1': 0.0, 'g2': 24.0, 'g3': 73.0},
        'I': {'g1': 6.0, 'g2': 15.0, 'g3': 100.0},
        'J': {'g1': 16.0, 'g2': -9.0, 'g3': 0.0},
        'K': {'g1': 26.0, 'g2': 17.0, 'g3': 17.0},
        'L': {'g1': 62.0, 'g2': 43.0, 'g3': 0.0}
    }


@pytest.fixture()
def preferences_dummy():
    return [Preference(superior='G', inferior='F'), Preference(superior='F', inferior='E')]


@pytest.fixture()
def indifferences_dummy():
    return [Indifference(equal1='D', equal2='G')]


@pytest.fixture()
def criterions_dummy():
    return [Criterion(criterion_id='g1', gain=True, number_of_linear_segments=0), Criterion(criterion_id='g2', gain=True, number_of_linear_segments=0), Criterion(criterion_id='g3', gain=True, number_of_linear_segments=0)]


@pytest.fixture()
def predefined_criterions_dummy():
    return [Criterion(criterion_id='g1', gain=True, number_of_linear_segments=4), Criterion(criterion_id='g2', gain=True, number_of_linear_segments=4), Criterion(criterion_id='g3', gain=True, number_of_linear_segments=4)]


@pytest.fixture()
def positions_dummy():
    return [Position(alternative_id='A', worst_position=12, best_position=2)]


@pytest.fixture()
def intensities_dummy():
    return [Intensity(alternative_id_1='H', alternative_id_2='G', alternative_id_3='B', alternative_id_4='D', criteria=['g1', 'g2'])]


@pytest.fixture()
def resolved_inconsistencies_dummy():
    return [[[], [], [Position(alternative_id='A', worst_position=12, best_position=9, criteria=[])], []], [[Preference(superior='G', inferior='F', criteria=[])], [], [], []], [[], [Indifference(equal1='D', equal2='G', criteria=[])], [], []]]


@pytest.fixture()
def hasse_diagram_dict_dummy():
    return {'A': ['F', 'K'], 'C': ['J'], 'D': ['G'], 'F': ['E', 'J'], 'G': ['B', 'D', 'F', 'H', 'K'], 'I': ['B'], 'K': ['C'], 'L': ['J'], 'B': [], 'E': [], 'H': [], 'J': []}


@pytest.fixture()
def representative_value_function_dict_dummy():
    return {'J': 0.0338, 'E': 0.1663, 'C': 0.196, 'F': 0.3015, 'K': 0.3311, 'B': 0.3344, 'H': 0.3972, 'L': 0.3972, 'A': 0.5323, 'D': 0.5756, 'G': 0.5756, 'I': 0.6227}


@pytest.fixture()
def criterion_functions_dummy():
    return {'g1': [(0.0, 0.0102816), (2.0, 0.0132191), (6.0, 0.0190943), (7.0, 0.0), (9.0, 0.0235007), (16.0, 0.0337822), (18.0, 0.0606056), (25.0, 0.154487), (26.0, 0.167899), (35.0, 0.288604), (39.0, 0.342251), (62.0, 0.342251)], 'g2': [(2.0, 0.0), (9.0, 0.0), (14.666666666666664, 0.0), (15.0, 0.000773555), (17.0, 0.00541489), (24.0, 0.0216596), (30.0, 0.0355836), (38.33333333333333, 0.0549224), (40.0, 0.0549224), (43.0, 0.0549224), (55.0, 0.0549224), (62.0, 0.0549224)], 'g3': [(0.0, 0.0), (12.0, 0.11142), (14.0, 0.12999), (17.0, 0.157845), (25.0, 0.232125), (33.33333333333333, 0.309501), (44.0, 0.309501), (66.66666666666666, 0.309501), (68.0, 0.321234), (73.0, 0.365232), (88.0, 0.497229), (100.0, 0.602826)]}


@pytest.fixture()
def predefined_hasse_diagram_dict_dummy():
    return {'A': ['F', 'K'], 'C': ['J'], 'D': ['G'], 'F': ['E', 'J'], 'G': ['B', 'D', 'F', 'H', 'K'], 'I': ['B'], 'K': ['C'], 'L': ['J'], 'B': [], 'E': [], 'H': [], 'J': []}


def test_get_hasse_diagram_dict(
        performance_table_dict_dummy,
        preferences_dummy,
        indifferences_dummy,
        criterions_dummy,
        positions_dummy,
        intensities_dummy,
        hasse_diagram_dict_dummy
):
    solver = Solver(show_logs=True)

    hasse_diagram_list = solver.get_hasse_diagram_dict(
        performance_table_dict_dummy,
        preferences_dummy,
        indifferences_dummy,
        criterions_dummy,
        positions_dummy,
        intensities_dummy
    )

    assert hasse_diagram_list == hasse_diagram_dict_dummy


def test_get_representative_value_function_dict(
        performance_table_dict_dummy,
        preferences_dummy,
        indifferences_dummy,
        predefined_criterions_dummy,
        criterions_dummy,
        positions_dummy,
        intensities_dummy,
        representative_value_function_dict_dummy,
        criterion_functions_dummy,
        resolved_inconsistencies_dummy
):
    solver = Solver(show_logs=True)

    try:
        representative_value_function_dict, criterion_functions, sampler_metrics = solver.get_representative_value_function_dict(
            performance_table_dict_dummy,
            preferences_dummy,
            indifferences_dummy,
            predefined_criterions_dummy,
            positions_dummy,
            intensities_dummy,
            'files/polyrun-1.1.0-jar-with-dependencies.jar',
            '10'
        )

        assert representative_value_function_dict == representative_value_function_dict_dummy
        assert criterion_functions == criterion_functions_dummy

    except Inconsistency as e:
        resolved_inconsistencies = e.data

        assert resolved_inconsistencies == resolved_inconsistencies_dummy


def test_predefined_get_hasse_diagram_dict(
        performance_table_dict_dummy,
        preferences_dummy,
        indifferences_dummy,
        predefined_criterions_dummy,
        positions_dummy,
        predefined_hasse_diagram_dict_dummy
):
    solver = Solver(show_logs=True)

    hasse_diagram_list = solver.get_hasse_diagram_dict(
        performance_table_dict_dummy,
        preferences_dummy,
        indifferences_dummy,
        predefined_criterions_dummy,
        positions_dummy
    )

    assert hasse_diagram_list == predefined_hasse_diagram_dict_dummy



from xmcda.criteria import Criteria
from xmcda.XMCDA import XMCDA
import csv
import _io
from typing import List, Dict

from .utils.parser_utils import ParserUtils
from .dataclasses import Criterion


class Parser:
    @staticmethod
    def get_performance_table_dict_csv(csvfile: _io.TextIOWrapper) -> Dict[str, Dict[str, float]]:
        """
        Method responsible for getting dict of performances from CSV file

        :param csvfile: python file object of csv file

        :return: Dictionary of performances
        """
        csv_reader = csv.reader(csvfile, delimiter=';')
        gains: List[str] = next(csv_reader)[1:]  # skip gains row
        criteria_ids: List[str] = next(csv_reader)[1:]

        performance_table_list: List[List[float]] = []
        alternative_ids: List[str] = []
        for row in csv_reader:
            performance_list: List[float] = [float(value) for value in row[1:]]
            alternative_id: str = [value for value in row[:1]][0]

            performance_table_list.append(performance_list)
            alternative_ids.append(alternative_id)

        result = {}
        for i in range(len(alternative_ids)):
            result[alternative_ids[i]] = {criteria_ids[j]: performance_table_list[i][j] for j in range(len(criteria_ids))}

        return result

    @staticmethod
    def get_criterion_list_csv(csvfile: _io.TextIOWrapper) -> List[Criterion]:
        """
        Method responsible for getting list of criteria

        :param csvfile: Content of file

        :return: List of criteria ex. ['g1', 'g2', 'g3']
        """
        csv_reader = csv.reader(csvfile, delimiter=';')

        gains: List[str] = next(csv_reader)[1:]
        criteria_ids: List[str] = next(csv_reader)[1:]

        criteria_objects = []
        for i in range(len(criteria_ids)):
            gain = True if gains[i].lower() == 'gain' else False
            criteria_objects.append(Criterion(criterion_id=criteria_ids[i], gain=gain, number_of_linear_segments=0))

        return criteria_objects

    @staticmethod
    def get_performance_table_dict_xmcda(xmcda_file: _io.TextIOWrapper) -> Dict[str, Dict[str, float]]:
        """
        Method responsible for getting dictionary representing performance table

        :param xmcda_file: XMCDA file

        :return: Dictionary representing performance table
            ex. {'id1': {'g1': 31.6, 'g2': 6.6, 'c3': 7.2}, 'id2': {'g1': 1.5, 'g2': 14.2, 'c3': 10.0}
        """
        xmcda_data = ParserUtils.load_xmcda(xmcda_file)
        performance_table_dict = {}
        for alternative in xmcda_data.alternatives:
            performance_list = {}
            for criterion in xmcda_data.criteria:
                performance_list[criterion.id] = xmcda_data.performance_tables[0][alternative][criterion]
            performance_table_dict[alternative.id] = performance_list

        return performance_table_dict

    @staticmethod
    def get_alternative_dict_xmcda(xmcda_file: _io.TextIOWrapper) -> Dict[str, str]:
        """
        Method responsible for getting dictionary of alternative

        :param xmcda_file: XMCDA file

        :return: Dictionary of alternatives ex. {'id1': 'Alternative1', 'id2': 'Alternative2', 'id3': 'Alternative3'}
        """
        xmcda_data = ParserUtils.load_xmcda(xmcda_file)
        alternatives = {}
        for alternative in xmcda_data.alternatives:
            alternatives[alternative.id] = alternative.name

        return alternatives

    @staticmethod
    def get_criterion_dict_xmcda(xmcda_file: _io.TextIOWrapper) -> Dict[str, Criterion]:
        """
        Method responsible for getting dictionary of criteria

        :param xmcda_file: XMCDA file

        :return: Dictionary of Criterion objects ex. ['id1': Criterion1,'id2': Criterion2,'id3': Criterion3]
        """
        xmcda_data = ParserUtils.load_xmcda(xmcda_file)
        criterion_dict = {}
        for criterion in xmcda_data.criteria:
            criterion_dict[criterion.id] = Criterion(criterion_id=criterion.name,
                                                     gain=(1 if criterion.id[0] == 'g' else 0),
                                                     number_of_linear_segments=0)

        return criterion_dict



import pytest
from src.utagmsengine.dataclasses import Criterion
from src.utagmsengine.parser import Parser
from typing import List, Dict
import os


@pytest.fixture()
def performance_table_dict_dummy():
    return {
        'A': {'g1': 26.0, 'g2': 40.0, 'g3': 44.0},
        'B': {'g1': 2.0, 'g2': 2.0, 'g3': 68.0},
        'C': {'g1': 18.0, 'g2': 17.0, 'g3': 14.0},
        'D': {'g1': 35.0, 'g2': 62.0, 'g3': 25.0},
        'E': {'g1': 7.0, 'g2': 55.0, 'g3': 12.0},
        'F': {'g1': 25.0, 'g2': 30.0, 'g3': 12.0},
        'G': {'g1': 9.0, 'g2': 62.0, 'g3': 88.0},
        'H': {'g1': 0.0, 'g2': 24.0, 'g3': 73.0},
        'I': {'g1': 6.0, 'g2': 15.0, 'g3': 100.0},
        'J': {'g1': 16.0, 'g2': 9.0, 'g3': 0.0},
        'K': {'g1': 26.0, 'g2': 17.0, 'g3': 17.0},
        'L': {'g1': 62.0, 'g2': 43.0, 'g3': 0.0}
    }


@pytest.fixture()
def criteria_list_dummy():
    return [
        Criterion(criterion_id='g1', gain=True, number_of_linear_segments=0),
        Criterion(criterion_id='g2', gain=True, number_of_linear_segments=0),
        Criterion(criterion_id='g3', gain=True, number_of_linear_segments=0)
    ]


# @pytest.fixture()
# def performance_table_list_dummy():
#     return [[26.0, 40.0, 44.0],
#             [2.0, 2.0, 68.0],
#             [18.0, 17.0, 14.0],
#             [35.0, 62.0, 25.0],
#             [7.0, 55.0, 12.0],
#             [25.0, 30.0, 12.0],
#             [9.0, 62.0, 88.0],
#             [0.0, 24.0, 73.0],
#             [6.0, 15.0, 100.0],
#             [16.0, 9.0, 0.0],
#             [26.0, 17.0, 17.0],
#             [62.0, 43.0, 0.0]]
#
#
# @pytest.fixture()
# def alternatives_id_list_dummy():
#     return ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L']
#
#
# @pytest.fixture()
# def criteria_list_dummy():
#     return [1, 1, 1]


# def test_get_performance_table_list_xml(performance_table_list_dummy):
#     parser = Parser()
#     performance_table_list: List[List[float]] = parser.get_performance_table_list_xml('performance_table.xml')
#
#     assert performance_table_list == performance_table_list_dummy
#
#
# def test_get_alternatives_id_list_xml(alternatives_id_list_dummy):
#     parser = Parser()
#     alternatives_id_list: List[str] = parser.get_alternatives_id_list_xml('alternatives.xml')
#
#     assert alternatives_id_list == alternatives_id_list_dummy
#
#
# def test_get_criteria_xml(criteria_list_dummy):
#     parser = Parser()
#
#     criteria_list: List[str] = parser.get_criteria_xml('performance_table.xml')
#
#     assert criteria_list == criteria_list_dummy


def test_get_performance_table_dict_csv(performance_table_dict_dummy):
    if os.environ.get('RUN_TESTS_LOCALLY', 'false').lower() == 'true':
        parser = Parser()
        with open('../tests/files/alternatives.csv', 'r') as csvfile:
            performance_table_dict: Dict[str, Dict[str, float]] = parser.get_performance_table_dict_csv(csvfile)

        assert performance_table_dict == performance_table_dict_dummy


def test_get_criterion_list_csv(criteria_list_dummy):
    if os.environ.get('RUN_TESTS_LOCALLY', 'false').lower() == 'true':
        parser = Parser()
        with open('../tests/files/alternatives.csv', 'r') as csvfile:
            criteria_list: List[Criterion] = parser.get_criterion_list_csv(csvfile)

        assert criteria_list == criteria_list_dummy




from pydantic import BaseModel, field_validator
from typing import List


class Preference(BaseModel):
    """
    Pydantic dataclass - Represents a preference between two options.
    Attributes:
        superior (str): The superior option.
        inferior (str): The inferior option.
    """
    superior: str
    inferior: str
    criteria: List[str] = []

    @field_validator("inferior")
    def check_different(cls, inferior, values):
        if "superior" in values.data and inferior == values.data["superior"]:
            raise ValueError("Superior and inferior options must be different.")
        return inferior

    @field_validator("criteria")
    def check_unique_criteria(cls, criteria):
        if criteria and len(set(criteria)) != len(criteria):
            raise ValueError("Criteria list must contain unique elements.")
        return criteria


class Indifference(BaseModel):
    """
    Pydantic dataclass - Represents an indifference criterion with two equal values.

    Attributes:
        equal1 (str): The first equal value.
        equal2 (str): The second equal value.
    """
    equal1: str
    equal2: str
    criteria: List[str] = []

    @field_validator("equal2")
    def check_different(cls, equal2, values):
        if "equal1" in values.data and equal2 == values.data["equal1"]:
            raise ValueError("First and second options must be different.")
        return equal2

    @field_validator("criteria")
    def check_unique_criteria(cls, criteria):
        if criteria and len(set(criteria)) != len(criteria):
            raise ValueError("Criteria list must contain unique elements.")
        return criteria


class Criterion(BaseModel):
    """
    Pydantic dataclass - Represents a decision criterion.

    Attributes:
        criterion_id (str): The unique identifier for the criterion.
        gain (bool): Whether the criterion represents a gain (True) or cost (False).
        number_of_linear_segments (int): The number of linear segments that the criterion has, if 0 then general function will be used
    """
    criterion_id: str
    gain: bool
    number_of_linear_segments: int

    @field_validator("number_of_linear_segments")
    def check_number_of_linear_segments(cls, number_of_linear_segments):
        if number_of_linear_segments < 0:
            raise ValueError("Number of linear segments can't be negative.")
        return number_of_linear_segments


class Position(BaseModel):
    alternative_id: str
    worst_position: int
    best_position: int
    criteria: List[str] = []

    @field_validator("worst_position")
    def check_worst_position(cls, worst_position):
        if worst_position < 0:
            raise ValueError("worst_position can't be negative.")
        return worst_position

    @field_validator("best_position")
    def check_max_position(cls, best_position):
        if best_position < 0:
            raise ValueError("best_position can't be negative.")
        return best_position

    @field_validator("criteria")
    def check_unique_criteria(cls, criteria):
        if criteria and len(set(criteria)) != len(criteria):
            raise ValueError("Criteria list must contain unique elements.")
        return criteria


class Intensity(BaseModel):
    """
    alternative_id_1 - alternative_id_2 >= alternative_id_3 - alternative_id_4 on given criteria
    """
    alternative_id_1: str
    alternative_id_2: str
    alternative_id_3: str
    alternative_id_4: str
    criteria: List[str] = []
    sign: str = '>='


class DataValidator:
    @staticmethod
    def validate_criteria(performance_table, criteria_list):
        """Validate whether Criterion IDs in performance_table and criteria_list match."""
        criteria_ids = {criterion.criterion_id for criterion in criteria_list}
        performance_table_ids = {key for item in performance_table.values() for key in item.keys()}

        if criteria_ids != performance_table_ids:
            raise ValueError("Criterion IDs in the list and the data dictionary do not match.")

    @staticmethod
    def validate_performance_table(performance_table):
        """Validate whether performance_table IDs are consistent"""
        keys_list = [set(inner_dict.keys()) for inner_dict in performance_table.values()]
        first_keys = keys_list[0]

        for keys in keys_list[1:]:
            if keys != first_keys:
                raise ValueError("Keys inside the inner dictionaries are not consistent.")

    @staticmethod
    def validate_positions(positions_list, performance_table):
        """Validate whether Alternative IDs in positions_list and performance_table match."""
        positions_ids = {position.alternative_id for position in positions_list}
        performance_table_ids = {key for key in performance_table.keys()}

        for position_id in positions_ids:
            if position_id not in performance_table_ids:
                raise ValueError("Alternative IDs in the Position list and the data dictionary do not match.")

        for position in positions_list:
            if position.worst_position < position.best_position:
                raise ValueError(f"worst_position can't be lower than best_position")

    @staticmethod
    def validate_preferences_indifferences_criteria(preferences, indifferences, positions, criteria):
        """Validate whether Alternative IDs in positions_list and performance_table match."""
        partial_criteria_list = []
        for preference in preferences:
            for criterion in preference.criteria:
                if criterion not in partial_criteria_list:
                    partial_criteria_list.append(criterion)

        for indifference in indifferences:
            for criterion in indifference.criteria:
                if criterion not in partial_criteria_list:
                    partial_criteria_list.append(criterion)

        for position in positions:
            for criterion in position.criteria:
                if criterion not in partial_criteria_list:
                    partial_criteria_list.append(criterion)

        criteria_list = [criterion.criterion_id for criterion in criteria]

        for criterion in partial_criteria_list:
            if criterion not in criteria_list:
                raise ValueError("Criteria name from partial criteria does not match name in Criterion")



from typing import Dict, List
import pytest
from pulp import LpProblem, value
from src.utagmsengine.utils.solver_utils import SolverUtils


@pytest.fixture()
def performance_table_list_dummy():
    return [[26.0, 40.0, 44.0],
            [2.0, 2.0, 68.0],
            [18.0, 17.0, 14.0],
            [35.0, 62.0, 25.0],
            [7.0, 55.0, 12.0],
            [25.0, 30.0, 12.0],
            [9.0, 62.0, 88.0],
            [0.0, 24.0, 73.0],
            [6.0, 15.0, 100.0],
            [16.0, 9.0, 0.0],
            [26.0, 17.0, 17.0],
            [62.0, 43.0, 0.0]]


@pytest.fixture()
def preferences_list_dummy():
    return [
        [6, 5, []],
        [5, 4, []]
    ]


@pytest.fixture()
def indifferences_list_dummy():
    return [
        [3, 6, []]
    ]


@pytest.fixture()
def criteria_list_dummy():
    return [1, 1, 1]


@pytest.fixture()
def alternatives_id_list_dummy():
    return ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L']


@pytest.fixture()
def problem_variable_values_dummy():
    return [0.5, 0.0, 0.5, 0.5, 0.0, 0.5, 0.5, 0.5, 0.0, 0.5, 0.0, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]


@pytest.fixture()
def necessary_dummy():
    return {'A': ['C', 'E', 'F', 'J', 'K'], 'C': ['J'], 'D': ['B', 'C', 'E', 'F', 'G', 'H', 'J', 'K'], 'F': ['E', 'J'], 'G': ['B', 'C', 'D', 'E', 'F', 'H', 'J', 'K'], 'I': ['B'], 'K': ['C', 'J'], 'L': ['J']}


@pytest.fixture()
def direct_relations_dummy():
    return {'A': ['F', 'K'], 'C': ['J'], 'D': ['G'], 'F': ['E', 'J'], 'G': ['B', 'D', 'F', 'H', 'K'], 'I': ['B'], 'K': ['C'], 'L': ['J']}


@pytest.fixture()
def alternatives_and_utilities_dict_dummy():
    return {'E': 0.0, 'C': 0.4667, 'F': 0.4667, 'J': 0.4667, 'K': 0.4667, 'L': 0.4667, 'B': 0.5333, 'H': 0.5333, 'I': 0.5333, 'A': 1.0, 'D': 1.0, 'G': 1.0}


@pytest.fixture()
def variables_and_values_dict_dummy():
    return {'epsilon': 0.18666667, 'u_0_0.0': 0.0, 'u_0_16.0': 0.46666667, 'u_0_18.0': 0.46666667, 'u_0_2.0': 0.0, 'u_0_25.0': 0.46666667, 'u_0_26.0': 0.46666667, 'u_0_35.0': 0.46666667, 'u_0_6.0': 0.0, 'u_0_62.0': 0.46666667, 'u_0_7.0': 0.0, 'u_0_9.0': 0.46666667, 'u_1_15.0': 0.0, 'u_1_17.0': 0.0, 'u_1_2.0': 0.0, 'u_1_24.0': 0.0, 'u_1_30.0': 0.0, 'u_1_40.0': 0.0, 'u_1_43.0': 0.0, 'u_1_55.0': 0.0, 'u_1_62.0': 0.0, 'u_1_9.0': 0.0, 'u_2_0.0': 0.0, 'u_2_100.0': 0.53333333, 'u_2_12.0': 0.0, 'u_2_14.0': 0.0, 'u_2_17.0': 0.0, 'u_2_25.0': 0.53333333, 'u_2_44.0': 0.53333333, 'u_2_68.0': 0.53333333, 'u_2_73.0': 0.53333333, 'u_2_88.0': 0.53333333}


@pytest.fixture()
def number_of_points_dummy():
    return [0, 0, 0]


@pytest.fixture()
def worst_best_positions_dummy():
    return [] #[[0, 6, 1]]


def test_create_variables_list_and_dict(performance_table_list_dummy):
    u_arr, u_arr_dict = SolverUtils.create_variables_list_and_dict(performance_table_list_dummy)

    assert len(u_arr) == 3
    assert len(u_arr_dict) == 3
    assert len(u_arr[0]) == 11
    assert len(u_arr[1]) == 10
    assert len(u_arr[2]) == 10
    assert u_arr[0][0].name == 'u_0_0.0'
    assert u_arr_dict[0][26.0].name == 'u_0_26.0'


def test_calculate_solved_problem(
    performance_table_list_dummy,
    preferences_list_dummy,
    indifferences_list_dummy,
    criteria_list_dummy,
    number_of_points_dummy,
    worst_best_positions_dummy,
    problem_variable_values_dummy
):
    problem: LpProblem = SolverUtils.calculate_solved_problem(
        performance_table_list=performance_table_list_dummy,
        preferences=preferences_list_dummy,
        indifferences=indifferences_list_dummy,
        criteria=criteria_list_dummy,
        worst_best_position=worst_best_positions_dummy,
        number_of_points=number_of_points_dummy,
        comprehensive_intensities=[],
        alternative_id_1=1,
        alternative_id_2=2
    )

    variable_values = []
    for var in problem.variables():
        variable_values.append(value(var))

    assert variable_values == problem_variable_values_dummy


def test_calculate_direct_relations(necessary_dummy, direct_relations_dummy):
    direct_relations: Dict[str, List[str]] = SolverUtils.calculate_direct_relations(necessary_dummy)

    assert direct_relations == direct_relations_dummy


def test_get_alternatives_and_utilities_dict(
        variables_and_values_dict_dummy,
        performance_table_list_dummy,
        alternatives_id_list_dummy,
        alternatives_and_utilities_dict_dummy
):
    alternatives_and_utilities_dict: Dict[str, float] = SolverUtils.get_alternatives_and_utilities_dict(
        variables_and_values_dict=variables_and_values_dict_dummy,
        performance_table_list=performance_table_list_dummy,
        alternatives_id_list=alternatives_id_list_dummy,
    )

    assert alternatives_and_utilities_dict == alternatives_and_utilities_dict_dummy





from xmcda.XMCDA import XMCDA
import _io

import os


class ParserUtils:

    @staticmethod
    def load_xmcda(file: _io.TextIOWrapper) -> XMCDA:
        """
        Private method responsible for loading XMCDA files from python file objects.

        :param file: XMCDA file

        :return: XMCDA
        """
        xmcda: XMCDA = XMCDA()
        xmcda_file: XMCDA = xmcda.load(file)
        return xmcda_file





from typing import List, Dict
from ..dataclasses import Preference, Indifference, Position, Intensity


class DataclassesUtils:

    @staticmethod
    def refine_performance_table_dict(
            performance_table_dict: Dict[str, Dict[str, float]]
    ) -> List[List[float]]:
        """
        Convert a dictionary of performance table values into a 2D list of floats.

        :param performance_table_dict:
        :return output_list:
        """
        output_list = []
        for key in performance_table_dict:
            inner_list = list(performance_table_dict[key].values())
            output_list.append(inner_list)

        return output_list

    @staticmethod
    def refine_preferences(
            performance_table_dict: Dict[str, Dict[str, float]],
            preferences
    ) -> List[List[int]]:
        """
        Convert a list of Preferences into a list of indices corresponding to alternatives.

        :param performance_table_dict:
        :param preferences:
        :return output:
        """
        output = []
        keys = list(performance_table_dict.keys())

        for preference in preferences:
            superior_index = keys.index(preference.superior)
            inferior_index = keys.index(preference.inferior)

            first = next(iter(performance_table_dict.values()))  # Get the first dictionary value
            criteria_index = []
            for criteria in preference.criteria:
                position = list(first.keys()).index(criteria)
                criteria_index.append(position)

            output.append([superior_index, inferior_index, criteria_index])

        return output

    @staticmethod
    def refine_indifferences(
            performance_table_dict: Dict[str, Dict[str, float]],
            indifferences
    ) -> List[List[int]]:
        """
        Convert a list of Indifferences into a list of indices corresponding to alternatives.

        :param performance_table_dict:
        :param indifferences:
        :return output:
        """
        output = []
        keys = list(performance_table_dict.keys())

        for indifference in indifferences:
            equal1_index = keys.index(indifference.equal1)
            equal2_index = keys.index(indifference.equal2)

            first = next(iter(performance_table_dict.values()))  # Get the first dictionary value
            criteria_index = []
            for criteria in indifference.criteria:
                position = list(first.keys()).index(criteria)
                criteria_index.append(position)

            output.append([equal1_index, equal2_index, criteria_index])

        return output

    @staticmethod
    def refine_gains(
            criterions
    ) -> List[bool]:
        """
        Extract gains/costs from a list of Criterion objects.

        :param criterions:

        :return output:
        """
        output = []

        for criterion in criterions:
            output.append(criterion.gain)

        return output

    @staticmethod
    def refine_linear_segments(
            criterions
    ) -> List[int]:
        """
        Extract number of linear segments from a list of Criterion objects.

        :param criterions:

        :return output:
        """
        output = []

        for criterion in criterions:
            output.append(criterion.number_of_linear_segments)

        return output

    @staticmethod
    def refine_positions(
            positions,
            performance_table_dict
    ) -> List[List[int]]:
        """
        Refined list[Positions] to [[alternative_ID, worst_position, best_position], ...] format

        :param positions:
        :param performance_table_dict:

        :return output:
        """
        output = []
        tmp = {}

        for i, key in enumerate(performance_table_dict.keys()):
            tmp[key] = i

        for position in positions:
            first = next(iter(performance_table_dict.values()))  # Get the first dictionary value
            criteria_index = []
            for criteria in position.criteria:
                index = list(first.keys()).index(criteria)
                criteria_index.append(index)

            output.append([tmp[position.alternative_id], position.worst_position, position.best_position, criteria_index])

        return output

    @staticmethod
    def refine_intensities(
            intensities,
            performance_table_dict
    ) -> List[List[int]]:
        """
        Refine into [[1,[],3,[],4,[],5,[],'>']], meaning 1-3 > 4-5

        :param intensities:
        :param performance_table_dict:

        :return output:
        """
        output = []
        tmp = {}

        for i, key in enumerate(performance_table_dict.keys()):
            tmp[key] = i

        for intensity in intensities:
            first = next(iter(performance_table_dict.values()))  # Get the first dictionary value
            criteria_index = []
            for criteria in intensity.criteria:
                index = list(first.keys()).index(criteria)
                criteria_index.append(index)

            output.append(
                [
                    tmp[intensity.alternative_id_1],
                    criteria_index,
                    tmp[intensity.alternative_id_2],
                    criteria_index,
                    tmp[intensity.alternative_id_3],
                    criteria_index,
                    tmp[intensity.alternative_id_4],
                    criteria_index,
                    intensity.sign
                ]
            )

        return output

    @staticmethod
    def refine_resolved_inconsistencies(
            resolved_inconsistencies,
            performance_table_dict,
    ) -> List[List[int]]:
        """
        Refine from [[1,[],3,[],4,[],5,[],'>']], meaning 1-3 > 4-5 into dataclasses

        :param resolved_inconsistencies:
        :param performance_table_dict:

        :return output:
        """
        output = []
        alt_idx = {}
        crit_idx = {}

        for i, key in enumerate(performance_table_dict.keys()):
            alt_idx[i] = key

        for i, key in enumerate(performance_table_dict[alt_idx[0]]):
            crit_idx[i] = key

        for inconsistencies in resolved_inconsistencies:
            preferences = []
            indifferences = []
            positions = []
            intensities = []

            for preference in inconsistencies[0]:
                preferences.append(
                    Preference(superior=alt_idx[preference[0]], inferior=alt_idx[preference[1]], criteria=[crit_idx[idx] for idx in preference[2]])
                )

            for indifference in inconsistencies[1]:
                indifferences.append(
                    Indifference(equal1=alt_idx[indifference[0]], equal2=alt_idx[indifference[1]], criteria=[crit_idx[idx] for idx in indifference[2]])
                )

            for position in inconsistencies[2]:
                positions.append(
                    Position(alternative_id=alt_idx[position[0]], worst_position=position[1], best_position=position[2], criteria=[crit_idx[idx] for idx in position[3]])
                )

            for intensity in inconsistencies[3]:
                intensities.append(
                    Intensity(
                        alternative_id_1=alt_idx[intensity[0]],
                        alternative_id_2=alt_idx[intensity[2]],
                        alternative_id_3=alt_idx[intensity[4]],
                        alternative_id_4=alt_idx[intensity[6]],
                        criteria=[crit_idx[idx] for idx in intensity[1]],
                        sign=intensity[8]
                    )
                )

            output.append(
                [
                    preferences,
                    indifferences,
                    positions,
                    intensities
                 ]
            )

        return output[:-1]
